/*
 * RCache - A collection of simple reference-based cache implementations.
 * Copyright (C) 2007  Rodrigo Ruiz
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301 USA
 *
 * Alternatively, the contents of this file may be used under the terms
 * of the Apache 2.0 license (the "Apache License"), in which case its
 * provisions are applicable instead of those above. If you wish to allow use
 * of your version of this file only* under the terms of the Apache License
 * and not to allow others to use your version of this file under the LGPL,
 * indicate your decision by* deleting the provisions above and replace them
 * with the notice and other provisions required by the Apache License. If
 * you do not delete the provisions above, a recipient may use your version of
 * this file under either the LGPL or the Apache License.
 */
package net.sourceforge.rcache;

import java.lang.ref.ReferenceQueue;

/**
 * <p>Memory-Sensitive Cache Based on SoftReferences</p>
 *
 * <p>Entries in a SoftCache can be garbage collected if, and only if, the JVM
 * considers that more memory is needed. This means that all SoftCache instances
 * will be effectively cleared before an OutOfMemory error is thrown.</p>
 *
 * <p>For certain applications, this class may be too much memory-consuming. If
 * the programmer only needs to enforce that a single instance is used for a
 * given key, and it is not important to keep this instance afterwards,
 * {@link WeakCache} may be a better option.</p>
 *
 * @see java.lang.ref.SoftReference
 *
 * @author Rodrigo Ruiz
 * @param <K> the type of keys maintained by this cache
 * @param <V> the type of cached values
 */
public class SoftCache<K, V> extends BaseCache<K, V> {

  /**
   * <p>Default constructor.</p>
   *
   * <p>Instances created with this constructor will have a small initial
   * capacity and a highly restricted concurrency level.</p>
   *
   * <p>For concurrent maps, resizing is a costly operation. Whenever possible,
   * it is advised to specify the expected maximum size, in order to
   * minimise the number of resizes.</p>
   */
  public SoftCache() {
    super(DEFAULT_INITIAL_CAPACITY, DEFAULT_LOAD_FACTOR, DEFAULT_CONCURRENCY_LEVEL);
  }

  /**
   * <p>Creates an instance with a custom initial capacity.</p>
   *
   * @param initialCapacity The map initial capacity
   */
  public SoftCache(int initialCapacity) {
    super(initialCapacity, DEFAULT_LOAD_FACTOR, DEFAULT_CONCURRENCY_LEVEL);
  }

  /**
   * <p>Creates an instance with a custom initial capacity and load factor.</p>
   *
   * @param initialCapacity The map initial capacity
   * @param loadFactor      The map load factor
   */
  public SoftCache(int initialCapacity, float loadFactor) {
    super(initialCapacity, loadFactor, DEFAULT_CONCURRENCY_LEVEL);
  }

  /**
   * <p>Creates an instance with custom capacity and concurrency level.</p>
   *
   * <p>The concurrency level is used as a hint to select the internal Map
   * implementation:</p>
   *
   * <ul>
   *   <li>&lt; 0 : Do not care about concurrency.
   *                A {@link java.util.HashMap} will be used.</li>
   *   <li>= 0    : No concurrent access allowed.
   *                A {@link java.util.Hashtable} will be selected.</li>
   *   <li>&gt; 0 : A {@link java.util.concurrent.ConcurrentHashMap} will be
   *                created with the specified concurrency level.</li>
   * </ul>
   *
   * @param initialCapacity  The map initial capacity
   * @param loadFactor       The map load factor
   * @param concurrencyLevel The concurrency level.
   */
  public SoftCache(int initialCapacity, float loadFactor, int concurrencyLevel) {
    super(initialCapacity, loadFactor, concurrencyLevel);
  }

  /**
   * {@inheritDoc}
   */
  @Override protected final KeyedReference<K, V> createRef(K key, V value,
    ReferenceQueue<V> queue) {
    return new SoftKeyedReference<K, V>(key, value, queue);
  }
}
