/*
 * RCache - A collection of simple reference-based cache implementations.
 * Copyright (C) 2007  Rodrigo Ruiz
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301 USA
 *
 * Alternatively, the contents of this file may be used under the terms
 * of the Apache 2.0 license (the "Apache License"), in which case its
 * provisions are applicable instead of those above. If you wish to allow use
 * of your version of this file only* under the terms of the Apache License
 * and not to allow others to use your version of this file under the LGPL,
 * indicate your decision by* deleting the provisions above and replace them
 * with the notice and other provisions required by the Apache License. If
 * you do not delete the provisions above, a recipient may use your version of
 * this file under either the LGPL or the Apache License.
 */
package net.sourceforge.rcache.decorator;

import net.sourceforge.rcache.Cache;
import static net.sourceforge.rcache.Cache.Operation.*;

/**
 * Keeps track of the average execution time for each operation.
 *
 * @author Rodrigo Ruiz
 * @param <K> the type of keys maintained by this cache
 * @param <V> the type of cached values
 */
public final class CacheProfiler<K, V> implements Cache<K, V> {

  /**
   * The wrapped cache.
   */
  private final Cache<K, V> cache;

  /**
   * Profile stats for each operation.
   */
  private final Sequence[] seq = {
    new Sequence(),
    new Sequence(),
    new Sequence(),
    new Sequence()
  };

  /**
   * Creates an instance.
   *
   * @param cache  The cache instance to wrap
   */
  public CacheProfiler(Cache<K, V> cache) {
    this.cache = cache;
  }

  /**
   * {@inheritDoc}
   */
  public V get(Object key) {
    long t0 = System.nanoTime();
    try {
      return this.cache.get(key);
    } finally {
      long t1 = System.nanoTime();
      seq[GET.ordinal()].add(t1 - t0);
    }
  }

  /**
   * {@inheritDoc}
   */
  public V put(K key, V value) {
    long t0 = System.nanoTime();
    try {
      return this.cache.put(key, value);
    } finally {
      long t1 = System.nanoTime();
      seq[PUT.ordinal()].add(t1 - t0);
    }
  }

  /**
   * {@inheritDoc}
   */
  public V remove(Object key) {
    long t0 = System.nanoTime();
    try {
      return this.cache.remove(key);
    } finally {
      long t1 = System.nanoTime();
      seq[REMOVE.ordinal()].add(t1 - t0);
    }
  }

  /**
   * {@inheritDoc}
   */
  public void clear() {
    long t0 = System.nanoTime();
    try {
      this.cache.clear();
    } finally {
      long t1 = System.nanoTime();
      seq[CLEAR.ordinal()].add(t1 - t0);
    }
  }

  /**
   * Gets the average execution time of an operation.
   *
   * @param op The operation
   * @return The operation average execution time
   */
  public double getAverage(Operation op) {
    return seq[op.ordinal()].getAverage();
  }

  /**
   * Gets the average execution time of all this cache operations.
   *
   * @return An array containing the average execution times in the same
   *         order as the elements of the {@link
   *         net.sourceforge.rcache.Cache.Operation Operation} enum
   */
  public double[] getAverages() {
    int i = 0;
    return new double[] {
      seq[i++].getAverage(),
      seq[i++].getAverage(),
      seq[i++].getAverage(),
      seq[i++].getAverage()
    };
  }

  /**
   * Gets the number of registered invocations to the specified Cache operation.
   *
   * @param op Operation
   * @return Number of tracked invocations
   */
  public long getNum(Operation op) {
    return seq[op.ordinal()].getNum();
  }

  /**
   * Gets the number of invocations to each Cache operation.
   *
   * @return An array containing the invocation tracking
   */
  public long[] getNums() {
    int i = 0;
    return new long[] {
      seq[i++].getNum(),
      seq[i++].getNum(),
      seq[i++].getNum(),
      seq[i++].getNum()
    };
  }

  /**
   * Holder for average execution times.
   */
  static class Sequence {
    /** Time sum. */
    private long sum;

    /** Call counter. */
    private long num;
    /**
     * Adds a single execution time.
     *
     * @param t The execution time
     */
    public synchronized void add(long t) {
      sum += t;
      num++;
    }
    /**
     * Gets the average execution time.
     *
     * @return The average execution time
     */
    public synchronized double getAverage() {
      return (num == 0L) ? 0.0 : (double)sum / (double)num;
    }
    /**
     * Gets the number of executions.
     *
     * @return The number of executions
     */
    public synchronized long getNum() {
      return num;
    }
  }
}
